require 'time'
require 'cgi'
require 'fileutils'
require 'base64'

module Fuel
  module CLI
    class Gerrit < Base
      include GerritCommon

      package_name "gerrit"

      headers "Content-Type" => "application/json"
      digest_auth Config.get('gerrit', 'user'), Config.get64('gerrit', 'password')
      parser GerritJsonParser

      desc "status", "Show a brief overview of the current gerrit status"
      def status
        change_id = fetch_change_id_or_fail

        result = self.class.get(changes_endpoint(change_id, "?o=CURRENT_REVISION")).parsed_response
        current_rev = result['current_revision']

        say result["subject"], :bold
        table = [["Owner", result["owner"]["name"]]]
        table << ["Updated", format_time(result["updated"])]
        table << ["Status", result["status"]]
        table << ["Ref", gerrit_ref(result)]
        print_table table

        result = fetch_comments(change_id, current_rev)
        print_comments_info(result)
        puts
        result = self.class.get(reviewers_endpoint(change_id)).parsed_response
        table = result.map { |row| format_row(row) }
        table.unshift(["Reviewer", "CR", "QA", "V"])
        print_table table
      end

      desc "open", "Open Gerrit in your browser"
      def open
        `open #{gerrit_change_url}`
      end

      desc "review-add USER[, USER, ...]", "Add reviewer(s) to the gerrit"
      def review_add(*emails_or_aliases)
        change_id = fetch_change_id_or_fail
        emails_or_aliases.each do |value|
          emails = Array(to_email(value))
          if emails.empty?
            say "No email found for alias #{value}; reviewer not added", :yellow
            next
          end
          emails.each do |email|
            result = self.class.post(reviewers_endpoint(change_id), body: { reviewer: email }.to_json).parsed_response
          end
        end
      end

      desc "alias EMAIL [ALIAS]", "Add an alias or show all existing aliases for the given email"
      def alias(email, alias_name = nil)
        if alias_name
          Config.set('gerrit', 'aliases', alias_name, email)
        else
          aliases = Config.get('gerrit', 'aliases') || {}
          aliases_found = aliases.map { |k, v| v == email ? k : nil }.compact
          if aliases_found.empty?
            say "No aliases created for #{email}", :yellow
          else
            say "Aliases for #{email}: #{aliases_found.join("; ")}"
          end
        end
      end

      desc "auth USER", "Store authentication information for Gerrit"
      def auth(user)
        set_credentials(user, ['gerrit', 'user'], ['gerrit', 'password'])
      end

      desc "submit", "Submit the current change"
      def submit
        change_id = fetch_change_id_or_fail
        result = self.class.post(changes_endpoint(change_id, 'submit'), body: { wait_for_merge: true }.to_json)
      end

      option :message, :aliases => '-m'
      desc "abandon", "Abandon the current change"
      def abandon
        change_id = fetch_change_id_or_fail
        params = options['message'] ? { body: { message: options['message'] }.to_json } : {}
        result = self.class.post(changes_endpoint(change_id, 'abandon'), params).parsed_response
      end

      option :message, :aliases => '-m'
      desc "restore", "Restore the current change"
      def restore
        change_id = fetch_change_id_or_fail
        params = options['message'] ? { body: { message: options['message'] }.to_json } : {}
        result = self.class.post(changes_endpoint(change_id, 'restore'), params).parsed_response
      end

      desc "rebase", "Rebase the current change"
      def rebase
        change_id = fetch_change_id_or_fail
        result = self.class.post(changes_endpoint(change_id, "rebase"))
        if result.response.code == "409"
          say "Rebase failed, conflicts found", :red
        else
          say "Rebased successfully", :green
        end
      end

      desc "checkout USER", "Pulls the latest patchset by the given user for local QA"
      def checkout(email_or_alias)
        email = to_email(email_or_alias)
        say "Group aliases are not supported", :red and return if email.is_a?(Array)
        result = self.class.get(changes_endpoint(nil, "?q=status:open+owner:#{email}&n=1&o=CURRENT_REVISION&o=DOWNLOAD_COMMANDS")).parsed_response
        say "No open changes found for #{email_or_alias}", :yellow and return if result.empty?
        current_rev = result[0]['current_revision']
        command = result[0]['revisions'][current_rev]['fetch']['ssh']['commands']['Checkout']
        `#{command}`
      end

      option :message, :aliases => '-m'
      desc "qa VOTE", "Put a QA mark on the patch currently checked out"
      def qa(vote)
        submit_review('QA-Test-Passed', vote)
      end

      option :message, :aliases => '-m'
      desc "cr VOTE", "Put a Code Review mark on the patch currently checked out"
      def cr(vote)
        submit_review('Code-Review', vote)
      end

      option :message, :aliases => '-m'
      desc "verified VOTE", "Put a Verified mark on the patch currently checked out"
      def verified(vote)
        submit_review('Verified', vote)
      end

      desc "comments FILE", "View the file with comments"
      def comments(file)
        comments, lines = comments_for_file(file)
        extra_lines = 0
        comments.each do |c|
          line = c['line'] + extra_lines
          comment_lines = format_comment(c)
          lines.insert(line, *comment_lines)
          extra_lines += comment_lines.size
        end
        exec("printf #{lines.join('\n').inspect} | less -F -X -R")
      end

      desc "respond FILE", "Look through the comments and respond to them"
      def respond(file)
        comments, lines, change_id, revision_id = comments_for_file(file)
        commented_on = {}
        comments_for_lines = comments.inject({}) do |h, c|
          h[c['line']] ||= []
          h[c['line']] << c
          h
        end
        comments_to_post = comments_for_lines.map do |line, cc|
          next if cc.all? { |c| c['author']['username'] == gerrit_user }
          system 'tput smcup'

          line_in_array = line - 1
          first_line = [0, line_in_array - 4].max
          if first_line == line_in_array
            say "Line #{line}:", :bold
          else
            say "Lines #{first_line + 1}-#{line}:", :bold
          end
          say lines[first_line..line_in_array].join("\n")
          cc.each do |c|
            say format_comment(c).join("\n")
          end
          reply = ask "Reply on this line (press enter to skip):"
          system 'tput rmcup'
          next if reply.strip.empty?
          { line: line, message: reply }
        end.compact
        say 'No comments to respond to!', :green and return if comments_to_post.empty?
        result = self.class.post(revisions_endpoint(change_id, revision_id, 'review'), body: { comments: { file => comments_to_post } }.to_json).parsed_response
      end

      no_commands do
        def gerrit_change_url
          change_id = fetch_change_id_or_fail
          result = self.class.get(changes_endpoint(change_id)).parsed_response
          "#{gerrit_url}/#{result['_number']}"
        end
      end

      private

      def format_comment(c)
        msg = c['message'].gsub("\n\n", "\n")
        date = format_time(c['updated'])
        author = set_color(c['author']['name'], :yellow)
        prefix = "[#{date}] #{author}: "
        comment_lines = msg.split("\n")
        comment_lines[0].prepend prefix
        comment_lines[1..-1].each do |cline|
          cline.prepend " " * prefix.size
        end
        comment_lines
      end

      def submit_review(label, vote)
        change_id = fetch_change_id_or_fail
        result = self.class.get(changes_endpoint(change_id, "?o=CURRENT_REVISION")).parsed_response
        current_rev = result['current_revision']
        body = { labels: { label => vote.to_i } }
        message = options['message'] || run_editor.gsub("\n", "\n\n")

        body[:message] = message
        result = self.class.post(revisions_endpoint(change_id, current_rev, 'review'), body: body.to_json).parsed_response
      end

      def to_email(value)
        is_email?(value) ? value : emails_for_alias(value)
      end

      def is_email?(value)
        !!value.index('@')
      end

      def emails_for_alias(alias_name)
        aliases = Config.get('gerrit', 'aliases', alias_name)
      end

      def format_row(row)
        result = [row["name"]]
        approvals = row["approvals"]

        result << format_review(approvals && approvals["Code-Review"])
        result << format_verified(approvals && approvals["QA-Test-Passed"])
        result << format_verified(approvals && approvals["Verified"])
        result
      end

      def format_review(cell)
        characters = { "0" => " ", "-1" => set_color("-1", :red), "+1" => set_color("+1", :green), "+2" => tick, "-2" => cross }
        characters[(cell || "").strip]
      end

      def format_verified(cell)
        characters = { "0" => " ", "-1" => cross, "+1" => tick }
        characters[(cell || "").strip]
      end

      def format_time(time)
        Time.parse(time + " UTC").getlocal.strftime("%b %-d %l:%M %p")
      end

      def tick
        set_color(" " + [10003].pack("U*"), :green)
      end

      def cross
        set_color(" " + [10008].pack("U*"), :red)
      end

      def revisions_endpoint(change_id, revision_id = nil, path = nil)
        url = "#{changes_endpoint(change_id)}revisions/"
        url += "#{revision_id}/" if revision_id
        url += path if path
        url
      end

      def reviewers_endpoint(change_id)
        changes_endpoint(change_id, "reviewers")
      end

      def run_editor
        path = File.expand_path('~/.fuel/MSG')
        FileUtils.touch(path)
        system((ENV['EDITOR'] || 'vi') + ' ' + path)
        content = File.read(path)
        FileUtils.rm(path)
        content
      end

      def fetch_comments(change_id, revision_id)
        return self.class.get(revisions_endpoint(change_id, revision_id, 'comments')).parsed_response
      end

      def comments_for_file(file)
        change_id = fetch_change_id_or_fail
        result = self.class.get(changes_endpoint(change_id, "detail?o=CURRENT_REVISION&o=CURRENT_FILES")).parsed_response
        current_rev = result['current_revision']

        content = self.class.get(revisions_endpoint(change_id, current_rev, "files/#{CGI::escape(file)}/content")).parsed_response
        content = Base64::decode64(content)
        lines = content.split("\n")
        comments = fetch_comments(change_id, current_rev)[file]
        comments.sort_by! { |c| [c['line'], Time.parse(c['updated'])] }
        [comments, lines, change_id, current_rev]
      end

      def gerrit_user
        Config.get('gerrit', 'user')
      end

      def print_comments_info(data)
        table = data.inject([]) do |memo, (file, comments)|
          memo << [file, set_color(plural(comments.count, 'comment'), :yellow)] if comments.count > 0
          memo
        end
        unless table.empty?
          puts
          say "Files with comments:"
          say "  (use \"gerrit comments FILE\" to see the comments)"
          print_table table, :indent => 8
        end
      end

      def plural(n, singular)
        if n == 1
          "1 #{singular}"
        else
          "#{n} #{singular}s"
        end
      end
    end
  end
end