module Fuel
  module CLI
    class Jenkins < Base
      include GerritCommon

      headers "Content-Type" => "application/json"
      digest_auth Config.get('gerrit', 'user'), Config.get64('gerrit', 'password')
      parser GerritJsonParser

      package_name "jenkins"

      desc "open", "Open the most recent Jenkins build for the current change"
      def open
        build_link = build_link_or_fail

        `open #{build_link}`
      end

      desc "deploy", "Deploy the latest patchset to your QA server"
      def deploy
        change_id = fetch_change_id_or_fail
        result = self.class.get(changes_endpoint(change_id, "?o=CURRENT_REVISION")).parsed_response
        current_rev = result['current_revision']

        client = create_client
        ref = gerrit_ref(result).sub('refs/', '')
        client.job.build("ui.fuel-deploy.gerrit", { "Environment" => "qa", "GerritRef" => ref })
      end

      desc "status", "View the status of the current build"
      def status
        details = build_details
        result = details['result']
        say 'Build is still running', :yellow and return unless result

        color = case result
        when "FAILURE", "ABORTED"
          :red
        when "SUCCESS"
          :green
        else
          :default
        end

        say result, color
        say 'Build artifacts are available via "jenkins artifacts"'
      end

      desc "artifacts", "Open build artifacts in the browser"
      def artifacts
        build_link = build_link_or_fail

        details = build_details
        say 'No artifacts available, build might be still running', :yellow and return if details['artifacts'].empty?

        `open #{build_link}artifact/`
      end

      desc "auth USER", "Store authentication information for Jenkins"
      def auth(user)
        set_credentials(user, ['jenkins', :username], ['jenkins', :password])
      end

      private

      def build_details
        build_link = build_link_or_fail
        build_number = build_link =~ /\/(\d+)\// && $1
        client = create_client

        client.job.get_build_details("ui.fuel-gerrit", build_number)
      end

      def create_client
        @client ||= Fuel::Util::JenkinsClient.new(Config.get('jenkins'))
      end

      def build_link_or_fail
        change_id = fetch_change_id_or_fail
        result = self.class.get(changes_endpoint(change_id, "detail")).parsed_response
        build_message = result['messages'].reverse.find { |m| m['author']['username'] == 'hbuilduser' }
        build_message &&= build_message['message']
        build_link = build_message && URI.extract(build_message)[0]
        raise Thor::Error, set_color('No Jenkins build found for this change', :red) unless build_link
        build_link
      end
    end
  end
end