module Fuel
  module CLI
    class Jira < Base
      package_name "jira"

      headers "Content-Type" => "application/json"
      basic_auth Config.get('jira', 'user'), Config.get64('jira', 'password')

      desc "status [JIRA]", "View a brief summary of the jira"
      def status(jira = nil)
        jira_id = jira || fetch_jira_or_fail
        result = self.class.get(issue_endpoint(jira_id)).parsed_response
        fields = result["fields"]
        say "#{jira_id}: #{fields['summary']}", :bold
        table = [['Type', fields['issuetype']['name']]]
        table << ['Status', fields['status']['name']]
        assignee = fields['assignee'] && fields['assignee']['displayName'] || 'Unassigned'
        table << ['Assignee', assignee]
        print_table table
      end

      desc "start [JIRA]", "Start progress on the jira"
      def start(jira = nil)
        transition('start progress', jira)
      end

      desc "resolve", "Resolve the current jira"
      def resolve
        transition('resolve issue')
      end

      desc "submit", "Put the current jira in review"
      def submit
        transition('submit', nil, fields: { get_field_id(fetch_jira_or_fail, 'review id') => Gerrit.new.gerrit_change_url })
      end

      desc "auth USER", "Store authentication information for JIRA"
      def auth(user)
        set_credentials(user, ['jira', 'user'], ['jira', 'password'])
      end

      private

      def transition(name, jira = nil, extra_body = {})
        jira_id = jira || fetch_jira_or_fail
        body = { transition: { id: get_transition_id(jira_id, name) } }.merge(extra_body)
        result = self.class.post(issue_endpoint(jira_id, "transitions"),
          body: body.to_json)
        say 'Jira status changed.', :green if result.code.to_i < 400
      end

      def jira_url
        Config.get('jira', 'url').chomp('/')
      end

      def get_field_id(jira_id, name)
        result = self.class.get(jira_endpoint("field")).parsed_response
        field = result.find { |r| r['name'].downcase == name.downcase }
        field && field['id']
      end

      def get_transition_id(jira_id, name)
        get_allowed_transitions(jira_id).each do |transition|
          return transition["id"] if transition["name"].downcase == name.downcase
        end
        nil
      end

      def get_allowed_transitions(jira_id)
        result = self.class.get(issue_endpoint(jira_id, "transitions")).parsed_response
        result["transitions"]
      end

      def fetch_jira_or_fail
        commit_message.split("\n")[2] or raise Thor::Error, set_color("No JIRA found at HEAD", :red)
      end

      def jira_endpoint(path = "")
        "#{jira_url}/rest/api/2/#{path}"
      end

      def issue_endpoint(id, path = "")
        jira_endpoint("issue/#{id}/#{path}")
      end
    end
  end
end
